unit lgpstruct;

interface

{$DEFINE SCANERR}

Uses Forms, Windows, SysUtils, Classes, Cappo;

Type

  TCreator = Array[0..11] of Byte;

  TProgressCallback = procedure(Cur,Max:Integer);

  TLGP_Header = packed record
    Creator:    TCreator;
    NumFiles:   Longint;
  end;

  TLGP_TableEntry = packed record
    Filename:   Array[0..19] of Char;
    EntryStart: Longint;
    Check1:     Byte;
    Check2:     Word;
  end;

  TLGP_FileHeader = packed record
    Filename:   Array[0..19] of Char;
    FileLen:    Longint;
  end;

  TLGP_Entry = record
    Filename:   String;
    TableStart: Integer; //filestart is TableStart +24
    FileLen:    Integer;
    Error:      Boolean;
  end;

  PLGP_Entry = ^TLGP_Entry;

  TLGPFile = class
    private
      FNumFiles:  Integer;
      FList:      TList;
      SList:      TList;
      FName,FCreat,FDesc:String;
      function GetEntry(Idx:Integer): TLGP_Entry;
      function GetSort(Idx:Integer): TLGP_Entry;
    public
      property Entry[Index:Integer]: TLGP_Entry read GetEntry; default;
      property Sorted[Index:Integer]: TLGP_Entry read GetSort;
      property NumFiles: Integer read FNumFiles;
      property Creator: String read FCreat;
      property Description: String read FDesc;
      property Filename: String read FName;
//      property Errors: Integer read FErr;
      Constructor CreateFromFile(FName:String;ReadHeaders:Boolean);
      Destructor Destroy; Override;
  end;

  TRawHeader = packed record
    Head:     TLGP_FileHeader;
    Offset:   Integer; //pos in file
    Data:     Pointer;
  end;

  PRawHeader = ^TRawHeader;

  PTable = ^TLGP_TableEntry;

  TRawLGPFile = class
    private
      Headers:    TList;
      Table:      TList;
      FNumHead:   Integer;
      FNumTable:  Integer;
      FCreat,FDesc:String;
      CRC:        PChar;
      Fil:        TFileStream;
      function GetHeader(Index:Integer): TRawHeader;
      function GetTbl(Index:INteger): TLGP_TableEntry;
      procedure SetTbl(Index:Integer; Tbl:TLGP_TableEntry);
    public
      property Description: String read FDesc;
      property Header[Index:Integer]: TRawHeader read GetHeader;
      property TableEntry[Index:Integer]: TLGP_TableEntry read GetTbl write SetTbl;
      property NumHeaders: Integer read FNumHead;
      property NumEntries: Integer read FNumTable;
      property Creator:String read FCreat;
      procedure AddHeader(NewHead:TRawHeader);
      procedure TruncateFrom(HeadNum:Integer);//zero based
      Constructor CreateFromFile(FName:String);
      Destructor Destroy; Override;
  end;

var
  Progress: TProgressCallback = nil;

procedure Extract(Strm:TStream;Ent:TLGP_Entry;Output:String);

Const
  LGPS = 1000;

implementation

procedure Extract(Strm:TStream;Ent:TLGP_Entry;Output:String);
var
FHead:      TLGP_FileHeader;
Fil:        TFileStream;
begin
If Ent.FileLen=-1 then begin
  Strm.Position := Ent.TableStart;
  Strm.ReadBuffer(FHead,Sizeof(FHead));
  Ent.FileLen := FHead.FileLen;
end else Strm.Position := Ent.TableStart + 24;
Fil := TFileStream.Create(Output,fmCreate or fmShareExclusive);
Fil.CopyFrom(Strm,Ent.FileLen);
Fil.Free;
end;

Constructor TLGPFile.CreateFromFile(FName:String;ReadHeaders:Boolean);
var
Ent:        PLGP_Entry;
LHead:      TLGP_Header;
Tbl:        TLGP_TableEntry;
FHead:      TLGP_FileHeader;
Fil:        TFileStream;
Err:        TFileStream;
Mx,Lrg,I,J: Integer;
S:          String;
Ch:         Char;
Done:       Boolean;
begin
Inherited;
FList := TList.Create; SList := TList.Create;
Self.FName := FName;
Fil := TFileStream.Create(FName,fmOpenRead or fmShareDenyNone);
try
Fil.ReadBuffer(LHead,Sizeof(LHead));
FNumFiles := LHead.NumFiles;
FCreat := ''; FDesc := '';
For I := 0 to 11 do
  If LHead.Creator[I] <> 0 then begin
    For J := I to 11 do
      FCreat := FCreat + Char(LHead.Creator[J]);
    Break;
  end;
Lrg := 0;
Mx := FNumFiles;
If ReadHeaders then Mx := 2*Mx;
For I := 1 to FNumFiles do begin
  Fil.ReadBuffer(Tbl,Sizeof(Tbl));
  New(Ent);
  Ent^.FileLen := -1;
  Ent^.Error := False;
  Ent^.Filename := StrPas(Tbl.Filename);
  Ent^.TableStart := Tbl.EntryStart;
//  If Tbl.Check1 <> 14 then Raise Exception.Create('Checkcode failure in LGP');
//  If Tbl.Check2 <> 0 then Raise Exception.Create('Checkcode failure in LGP');
  FList.Add(Ent); Done := False;
  For J := SList.Count-1 downto 0 do
    If PLGP_Entry(SList[J]).TableStart < Ent^.TableStart then begin
      SList.Insert(J+1,Ent); Done := True;
      Break;
    end;
  If Not Done then SList.Insert(0,Ent);
  Lrg := Max(Lrg,Tbl.EntryStart);
  If @Progress<>nil then Progress(I,Mx);
end;

{$IFDEF SCANERR}
If Not FileExists(ExtractFilePath(ParamStr(0))+'LGP.'+IntToStr(FNumFiles)) then begin
  Err := TFileStream.Create(ExtractFilePath(ParamStr(0))+'LGP.'+IntToStr(FNumFiles),fmCreate or fmShareExclusive);
  Err.CopyFrom(Fil,3602);
  Err.Free;
end;
{$ENDIF}

If ReadHeaders then
  For I := 0 to FList.Count-1 do begin
    Ent := FList[I];
    Fil.Position := Ent^.TableStart;
    Fil.ReadBuffer(FHead,Sizeof(FHead));
    S := StrPas(FHead.Filename);
    If Uppercase(S) <> Uppercase(Ent^.Filename) then Ent^.Error := True;
    Ent^.FileLen := FHead.FileLen;
    If Ent^.TableStart=Lrg then begin
      Fil.Position := Ent^.TableStart + Sizeof(FHead) + Ent^.FileLen;
      While Fil.Position < Fil.Size do begin
        Fil.ReadBuffer(Ch,1);
        FDesc := FDesc + Ch;
      end;
    end;
    If @Progress<>nil then Progress(FNumFiles+I,Mx);
  end;

finally
Fil.Free;
end;

end;

Destructor TLGPFile.Destroy;
var
Ent:      PLGP_Entry;
I:        Integer;
begin
For I := FList.Count-1 downto 0 do begin
  Ent := FList[I];
  Dispose(Ent);
end;
FList.Free;
SList.Free;
Inherited;
end;

function TLGPFile.GetEntry(Idx:Integer): TLGP_Entry;
var
Ent:      PLGP_Entry;
begin
Ent := FList[Idx];
Result := Ent^;
end;

function TLGPFile.GetSort(Idx:Integer): TLGP_Entry;
var
Ent:      PLGP_Entry;
begin
Ent := SList[Idx];
Result := Ent^;
end;

Constructor TRawLGPFile.CreateFromFile(FName:String);
var
I,Sml,J:    Integer;
LHead:      TLGP_Header;
Tbl:        PTable;
Raw:        PRawHeader;
Ch:         Char;
begin
Inherited;
Headers := TList.Create; Table := TList.Create; CRC := nil;
FNumTable := 0; FNumHead := 0;

Fil := TFileStream.Create(FName,fmOpenReadWrite or fmShareDenyWrite);
Fil.ReadBuffer(LHead,Sizeof(LHead));
For I := 0 to 11 do
  If LHead.Creator[I] <> 0 then begin
    For J := I to 11 do
      FCreat := FCreat + Char(LHead.Creator[J]);
    Break;
  end;
FNumTable := LHead.NumFiles;

Sml := Fil.Size;
For I := 0 to LHead.NumFiles-1 do begin
  New(Tbl);
  Fil.ReadBuffer(Tbl^,Sizeof(TLGP_TableEntry));
  Table.Add(Tbl);
  Sml := Min(Sml,Tbl^.EntryStart);
end;

If Sml > Fil.Position then begin
  GetMem(CRC,Sml-Fil.Position);
  Fil.ReadBuffer(CRC^,Sml-Fil.Position);
end;

While (Fil.Size - Fil.Position) > Sizeof(TLGP_FileHeader) do begin
  New(Raw);
  Raw^.Offset := Fil.Position;
  Raw^.Data := nil;
  Fil.ReadBuffer(Raw^.Head,Sizeof(TLGP_FileHeader));
  Fil.Position := Fil.Position + Raw^.Head.FileLen;
  Headers.Add(Raw);
end;

FNumHead := Headers.Count;
FDesc := '';
While Fil.Position < Fil.Size do begin
  Fil.ReadBuffer(Ch,1);
  FDesc := FDesc + Ch;
end;

end;

Destructor TRawLGPFile.Destroy;
var
Tbl:        PTable;
Raw:        PRawHeader;
I:          Integer;
begin
For I := Headers.Count-1 downto 0 do begin
  Raw := Headers[I];
  Dispose(Raw);
end;
For I := Table.Count-1 downto 0 do begin
  Tbl := Table[I];
  Dispose(Tbl);
end;

If CRC<>nil then Dispose(CRC);

Headers.Free; Table.Free; Fil.Free;
Inherited;
end;

procedure TRawLGPFile.AddHeader(NewHead:TRawHeader);
var
Raw:      PRawHeader;
Ch:       Char;
I:        Integer;
begin
Raw := Headers[Headers.Count-1];
Fil.Position := Raw^.Offset + Raw^.Head.FileLen + Sizeof(TLGP_FileHeader);
NewHead.Offset := Fil.Position;
Fil.WriteBuffer(NewHead.Head, Sizeof(TLGP_FileHeader));
Fil.WriteBuffer(NewHead.Data^, NewHead.Head.FileLen);
For I := 1 to Length(FDesc) do begin
  Ch := FDesc[I];
  Fil.WriteBuffer(Ch,1);
end;
New(Raw);
Raw^ := NewHead;
Headers.Add(Raw);
Inc(FNumHead);
end;

function TRawLGPFile.GetHeader(Index:Integer): TRawHeader;
var
Raw:    PRawHeader;
begin
Raw := Headers[Index];
Result := Raw^;
end;

function TRawLGPFile.GetTbl(Index:INteger): TLGP_TableEntry;
var
Tbl:        PTable;
begin
Tbl := Table[Index];
Result := Tbl^;
end;

procedure TRawLGPFile.SetTbl(Index:Integer; Tbl:TLGP_TableEntry);
var
ETbl:        PTable;
begin
ETbl := Table[Index];
ETbl^ := Tbl;
Fil.Position := Sizeof(TLGP_Header) + Sizeof(TLGP_TableEntry) * Index;
Fil.WriteBuffer(Tbl,Sizeof(Tbl));
end;

procedure TRawLGPFile.TruncateFrom(HeadNum:Integer);//zero based
var
I:  Integer;
Ch: Char;
Head:TRawHeader;
begin
Head := Header[HeadNum];
Fil.Size := Head.Offset + Length(FDesc);
Fil.Position := Head.Offset;
For I := 1 to Length(FDesc) do begin
  Ch := FDesc[I];
  Fil.WriteBuffer(Ch,1);
end;
For I := FNumHead-1 downto HeadNum do
  Headers.Delete(I);
FNumHead := HeadNum;
end;

end.
